#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from struct import pack as struct_pack

from neo4j.exceptions import (
    AuthError,
    Neo4jError,
    ServiceUnavailable,
    SessionExpired,
)
from neo4j.packstream import (
    UnpackableBuffer,
    Unpacker,
)

import logging
log = logging.getLogger("neo4j")


class MessageInbox:

    def __init__(self, s, on_error):
        self.on_error = on_error
        self._messages = self._yield_messages(s)

    def _yield_messages(self, sock):
        try:
            buffer = UnpackableBuffer()
            unpacker = Unpacker(buffer)
            chunk_size = 0
            while True:

                while chunk_size == 0:
                    # Determine the chunk size and skip noop
                    buffer.receive(sock, 2)
                    chunk_size = buffer.pop_u16()
                    if chunk_size == 0:
                        log.debug("[#%04X]  S: <NOOP>", sock.getsockname()[1])

                buffer.receive(sock, chunk_size + 2)
                chunk_size = buffer.pop_u16()

                if chunk_size == 0:
                    # chunk_size was the end marker for the message
                    size, tag = unpacker.unpack_structure_header()
                    fields = [unpacker.unpack() for _ in range(size)]
                    yield tag, fields
                    # Reset for new message
                    unpacker.reset()

        except OSError as error:
            self.on_error(error)

    def pop(self):
        return next(self._messages)


class Inbox(MessageInbox):

    def __next__(self):
        tag, fields = self.pop()
        if tag == b"\x71":
            return fields, None, None
        elif fields:
            return [], tag, fields[0]
        else:
            return [], tag, None


class Outbox:

    def __init__(self, capacity=8192, max_chunk_size=16384):
        self._max_chunk_size = max_chunk_size
        self._header = 0
        self._start = 2
        self._end = 2
        self._data = bytearray(capacity)

    def max_chunk_size(self):
        return self._max_chunk_size

    def clear(self):
        self._header = 0
        self._start = 2
        self._end = 2
        self._data[0:2] = b"\x00\x00"

    def write(self, b):
        to_write = len(b)
        max_chunk_size = self._max_chunk_size
        pos = 0
        while to_write > 0:
            chunk_size = self._end - self._start
            remaining = max_chunk_size - chunk_size
            if remaining == 0 or remaining < to_write <= max_chunk_size:
                self.chunk()
            else:
                wrote = min(to_write, remaining)
                new_end = self._end + wrote
                self._data[self._end:new_end] = b[pos:pos+wrote]
                self._end = new_end
                pos += wrote
                new_chunk_size = self._end - self._start
                self._data[self._header:(self._header + 2)] = struct_pack(">H", new_chunk_size)
                to_write -= wrote

    def chunk(self):
        self._header = self._end
        self._start = self._header + 2
        self._end = self._start
        self._data[self._header:self._start] = b"\x00\x00"

    def view(self):
        end = self._end
        chunk_size = end - self._start
        if chunk_size == 0:
            return memoryview(self._data[:self._header])
        else:
            return memoryview(self._data[:end])


class Response:
    """ Subscriber object for a full response (zero or
    more detail messages followed by one summary message).
    """

    def __init__(self, connection, **handlers):
        self.connection = connection
        self.handlers = handlers
        self.complete = False

    def on_records(self, records):
        """ Called when one or more RECORD messages have been received.
        """
        handler = self.handlers.get("on_records")
        if callable(handler):
            handler(records)

    def on_success(self, metadata):
        """ Called when a SUCCESS message has been received.
        """
        handler = self.handlers.get("on_success")
        if callable(handler):
            handler(metadata)

        if not metadata.get("has_more"):
            handler = self.handlers.get("on_summary")
            if callable(handler):
                handler()

    def on_failure(self, metadata):
        """ Called when a FAILURE message has been received.
        """
        try:
            self.connection.reset()
        except (SessionExpired, ServiceUnavailable):
            pass
        handler = self.handlers.get("on_failure")
        if callable(handler):
            handler(metadata)
        handler = self.handlers.get("on_summary")
        if callable(handler):
            handler()
        raise Neo4jError.hydrate(**metadata)

    def on_ignored(self, metadata=None):
        """ Called when an IGNORED message has been received.
        """
        handler = self.handlers.get("on_ignored")
        if callable(handler):
            handler(metadata)
        handler = self.handlers.get("on_summary")
        if callable(handler):
            handler()


class InitResponse(Response):

    def on_failure(self, metadata):
        code = metadata.get("code")
        message = metadata.get("message", "Connection initialisation failed")
        if code == "Neo.ClientError.Security.Unauthorized":
            raise AuthError(message)
        else:
            raise ServiceUnavailable(message)


class CommitResponse(Response):

    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest import TestCase

from neo4j.data import DataHydrator
from neo4j.graph import (
    Node,
    Path,
    Graph,
    Relationship,
)
from neo4j.packstream import Structure

# python -m pytest -s -v tests/unit/test_types.py


# Node


def test_can_create_node():
    g = Graph()
    gh = Graph.Hydrator(g)
    alice = gh.hydrate_node(1, {"Person"}, {"name": "Alice", "age": 33})
    assert isinstance(alice, Node)
    assert alice.labels == {"Person"}
    assert set(alice.keys()) == {"name", "age"}
    assert set(alice.values()) == {"Alice", 33}
    assert set(alice.items()) == {("name", "Alice"), ("age", 33)}
    assert alice.get("name") == "Alice"
    assert alice.get("age") == 33
    assert len(alice) == 2
    assert alice["name"] == "Alice"
    assert alice["age"] == 33
    assert "name" in alice
    assert "age" in alice
    assert set(iter(alice)) == {"name", "age"}


def test_null_properties():
    g = Graph()
    gh = Graph.Hydrator(g)
    stuff = gh.hydrate_node(1, (), {"good": ["puppies", "kittens"], "bad": None})
    assert isinstance(stuff, Node)
    assert set(stuff.keys()) == {"good"}
    assert stuff.get("good") == ["puppies", "kittens"]
    assert stuff.get("bad") is None
    assert len(stuff) == 1
    assert stuff["good"] == ["puppies", "kittens"]
    assert stuff["bad"] is None
    assert "good" in stuff
    assert "bad" not in stuff


def test_node_equality():
    g = Graph()
    node_1 = Node(g, 1234)
    node_2 = Node(g, 1234)
    node_3 = Node(g, 5678)
    assert node_1 == node_2
    assert node_1 != node_3
    assert node_1 != "this is not a node"


def test_node_hashing():
    g = Graph()
    node_1 = Node(g, 1234)
    node_2 = Node(g, 1234)
    node_3 = Node(g, 5678)
    assert hash(node_1) == hash(node_2)
    assert hash(node_1) != hash(node_3)


def test_node_repr():
    g = Graph()
    gh = Graph.Hydrator(g)
    alice = gh.hydrate_node(1, {"Person"}, {"name": "Alice"})
    assert repr(alice) == "<Node id=1 labels=frozenset({'Person'}) properties={'name': 'Alice'}>"


# Relationship


def test_can_create_relationship():
    g = Graph()
    gh = Graph.Hydrator(g)
    alice = gh.hydrate_node(1, {"Person"}, {"name": "Alice", "age": 33})
    bob = gh.hydrate_node(2, {"Person"}, {"name": "Bob", "age": 44})
    alice_knows_bob = gh.hydrate_relationship(1, alice.id, bob.id, "KNOWS", {"since": 1999})
    assert isinstance(alice_knows_bob, Relationship)
    assert alice_knows_bob.start_node == alice
    assert alice_knows_bob.type == "KNOWS"
    assert alice_knows_bob.end_node == bob
    assert set(alice_knows_bob.keys()) == {"since"}
    assert set(alice_knows_bob.values()) == {1999}
    assert set(alice_knows_bob.items()) == {("since", 1999)}
    assert alice_knows_bob.get("since") == 1999


def test_relationship_repr():
    g = Graph()
    gh = Graph.Hydrator(g)
    alice = gh.hydrate_node(1, {"Person"}, {"name": "Alice"})
    bob = gh.hydrate_node(2, {"Person"}, {"name": "Bob"})
    alice_knows_bob = gh.hydrate_relationship(1, alice.id, bob.id, "KNOWS", {"since": 1999})
    assert repr(alice_knows_bob) == "<Relationship id=1 nodes=(<Node id=1 labels=frozenset({'Person'}) properties={'name': 'Alice'}>, <Node id=2 labels=frozenset({'Person'}) properties={'name': 'Bob'}>) type='KNOWS' properties={'since': 1999}>"


# Path


def test_can_create_path():
    g = Graph()
    gh = Graph.Hydrator(g)
    alice = gh.hydrate_node(1, {"Person"}, {"name": "Alice", "age": 33})
    bob = gh.hydrate_node(2, {"Person"}, {"name": "Bob", "age": 44})
    carol = gh.hydrate_node(3, {"Person"}, {"name": "Carol", "age": 55})
    alice_knows_bob = gh.hydrate_relationship(1, alice.id, bob.id, "KNOWS", {"since": 1999})
    carol_dislikes_bob = gh.hydrate_relationship(2, carol.id, bob.id, "DISLIKES", {})
    path = Path(alice, alice_knows_bob, carol_dislikes_bob)
    assert isinstance(path, Path)
    assert path.start_node == alice
    assert path.end_node == carol
    assert path.nodes == (alice, bob, carol)
    assert path.relationships == (alice_knows_bob, carol_dislikes_bob)
    assert list(path) == [alice_knows_bob, carol_dislikes_bob]


def test_can_hydrate_path():
    g = Graph()
    gh = Graph.Hydrator(g)
    alice = gh.hydrate_node(1, {"Person"}, {"name": "Alice", "age": 33})
    bob = gh.hydrate_node(2, {"Person"}, {"name": "Bob", "age": 44})
    carol = gh.hydrate_node(3, {"Person"}, {"name": "Carol", "age": 55})
    r = [gh.hydrate_unbound_relationship(1, "KNOWS", {"since": 1999}),
         gh.hydrate_unbound_relationship(2, "DISLIKES", {})]
    path = gh.hydrate_path([alice, bob, carol], r, [1, 1, -2, 2])
    assert path.start_node == alice
    assert path.end_node == carol
    assert path.nodes == (alice, bob, carol)
    expected_alice_knows_bob = gh.hydrate_relationship(1, alice.id, bob.id, "KNOWS", {"since": 1999})
    expected_carol_dislikes_bob = gh.hydrate_relationship(2, carol.id, bob.id, "DISLIKES", {})
    assert path.relationships == (expected_alice_knows_bob, expected_carol_dislikes_bob)
    assert list(path) == [expected_alice_knows_bob, expected_carol_dislikes_bob]


def test_path_equality():
    g = Graph()
    gh = Graph.Hydrator(g)
    alice = gh.hydrate_node(1, {"Person"}, {"name": "Alice", "age": 33})
    bob = gh.hydrate_node(2, {"Person"}, {"name": "Bob", "age": 44})
    carol = gh.hydrate_node(3, {"Person"}, {"name": "Carol", "age": 55})
    alice_knows_bob = gh.hydrate_relationship(1, alice.id, bob.id, "KNOWS", {"since": 1999})
    carol_dislikes_bob = gh.hydrate_relationship(2, carol.id, bob.id, "DISLIKES", {})
    path_1 = Path(alice, alice_knows_bob, carol_dislikes_bob)
    path_2 = Path(alice, alice_knows_bob, carol_dislikes_bob)
    assert path_1 == path_2
    assert path_1 != "this is not a path"


def test_path_hashing():
    g = Graph()
    gh = Graph.Hydrator(g)
    alice = gh.hydrate_node(1, {"Person"}, {"name": "Alice", "age": 33})
    bob = gh.hydrate_node(2, {"Person"}, {"name": "Bob", "age": 44})
    carol = gh.hydrate_node(3, {"Person"}, {"name": "Carol", "age": 55})
    alice_knows_bob = gh.hydrate_relationship(1, alice.id, bob.id, "KNOWS", {"since": 1999})
    carol_dislikes_bob = gh.hydrate_relationship(2, carol.id, bob.id, "DISLIKES", {})
    path_1 = Path(alice, alice_knows_bob, carol_dislikes_bob)
    path_2 = Path(alice, alice_knows_bob, carol_dislikes_bob)
    assert hash(path_1) == hash(path_2)


def test_path_repr():
    g = Graph()
    gh = Graph.Hydrator(g)
    alice = gh.hydrate_node(1, {"Person"}, {"name": "Alice"})
    bob = gh.hydrate_node(2, {"Person"}, {"name": "Bob"})
    carol = gh.hydrate_node(3, {"Person"}, {"name": "Carol"})
    alice_knows_bob = gh.hydrate_relationship(1, alice.id, bob.id, "KNOWS", {"since": 1999})
    carol_dislikes_bob = gh.hydrate_relationship(2, carol.id, bob.id, "DISLIKES", {})
    path = Path(alice, alice_knows_bob, carol_dislikes_bob)
    assert repr(path) == "<Path start=<Node id=1 labels=frozenset({'Person'}) properties={'name': 'Alice'}> end=<Node id=3 labels=frozenset({'Person'}) properties={'name': 'Carol'}> size=2>"



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from math import isnan

import pytest


def test_null(cypher_eval):
    assert cypher_eval("RETURN null") is None


def test_boolean_true(cypher_eval):
    assert cypher_eval("RETURN true") is True


def test_boolean_false(cypher_eval):
    assert cypher_eval("RETURN false") is False


def test_integer(cypher_eval):
    assert cypher_eval("RETURN 123456789") == 123456789


def test_float(cypher_eval):
    assert cypher_eval("RETURN 3.1415926") == 3.1415926


def test_float_nan(cypher_eval):
    assert isnan(cypher_eval("WITH $x AS x RETURN x", x=float("NaN")))


def test_float_positive_infinity(cypher_eval):
    infinity = float("+Inf")
    assert cypher_eval("WITH $x AS x RETURN x", x=infinity) == infinity


def test_float_negative_infinity(cypher_eval):
    infinity = float("-Inf")
    assert cypher_eval("WITH $x AS x RETURN x", x=infinity) == infinity


def test_string(cypher_eval):
    assert cypher_eval("RETURN 'hello, world'") == "hello, world"


def test_bytes(cypher_eval):
    data = bytearray([0x00, 0x33, 0x66, 0x99, 0xCC, 0xFF])
    assert cypher_eval("CREATE (a {x:$x}) RETURN a.x", x=data) == data


def test_list(cypher_eval):
    data = ["one", "two", "three"]
    assert cypher_eval("WITH $x AS x RETURN x", x=data) == data


def test_map(cypher_eval):
    data = {"one": "eins", "two": "zwei", "three": "drei"}
    assert cypher_eval("WITH $x AS x RETURN x", x=data) == data


def test_non_string_map_keys(session):
    with pytest.raises(TypeError):
        _ = session.run("RETURN $x", x={1: 'eins', 2: 'zwei', 3: 'drei'})



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from uuid import uuid4


from neo4j.work.simple import (
    Query,
    TransactionError,
)
from neo4j.exceptions import (
    CypherSyntaxError,
    ClientError,
    TransientError,
)


def test_can_commit_transaction(session):

    tx = session.begin_transaction()

    # Create a node
    result = tx.run("CREATE (a) RETURN id(a)")
    record = next(iter(result))
    node_id = record[0]
    assert isinstance(node_id, int)

    # Update a property
    tx.run("MATCH (a) WHERE id(a) = $n "
           "SET a.foo = $foo", {"n": node_id, "foo": "bar"})

    tx.commit()

    # Check the property value
    result = session.run("MATCH (a) WHERE id(a) = $n "
                         "RETURN a.foo", {"n": node_id})
    record = next(iter(result))
    value = record[0]
    assert value == "bar"


def test_can_rollback_transaction(session):
    tx = session.begin_transaction()

    # Create a node
    result = tx.run("CREATE (a) RETURN id(a)")
    record = next(iter(result))
    node_id = record[0]
    assert isinstance(node_id, int)

    # Update a property
    tx.run("MATCH (a) WHERE id(a) = $n "
           "SET a.foo = $foo", {"n": node_id, "foo": "bar"})

    tx.rollback()

    # Check the property value
    result = session.run("MATCH (a) WHERE id(a) = $n "
                         "RETURN a.foo", {"n": node_id})
    assert len(list(result)) == 0


def test_can_commit_transaction_using_with_block(session):
    with session.begin_transaction() as tx:
        # Create a node
        result = tx.run("CREATE (a) RETURN id(a)")
        record = next(iter(result))
        node_id = record[0]
        assert isinstance(node_id, int)

        # Update a property
        tx.run("MATCH (a) WHERE id(a) = $n "
               "SET a.foo = $foo", {"n": node_id, "foo": "bar"})

        tx.commit()

    # Check the property value
    result = session.run("MATCH (a) WHERE id(a) = $n "
                         "RETURN a.foo", {"n": node_id})
    record = next(iter(result))
    value = record[0]
    assert value == "bar"


def test_can_rollback_transaction_using_with_block(session):
    with session.begin_transaction() as tx:
        # Create a node
        result = tx.run("CREATE (a) RETURN id(a)")
        record = next(iter(result))
        node_id = record[0]
        assert isinstance(node_id, int)

        # Update a property
        tx.run("MATCH (a) WHERE id(a) = $n "
               "SET a.foo = $foo", {"n": node_id, "foo": "bar"})
        tx.rollback()

    # Check the property value
    result = session.run("MATCH (a) WHERE id(a) = $n "
                         "RETURN a.foo", {"n": node_id})
    assert len(list(result)) == 0


def test_broken_transaction_should_not_break_session(session):
    with pytest.raises(CypherSyntaxError):
        with session.begin_transaction() as tx:
            tx.run("X")
    with session.begin_transaction() as tx:
        tx.run("RETURN 1")


def test_begin_transaction_time_out_metadata(session):
    with session.begin_transaction(timeout=2, metadata={"foo": "bar"}) as transaction:
        _ = transaction.run("RETURN 1").single()


def test_statement_object_not_supported(session):
    with session.begin_transaction() as tx:
        with pytest.raises(ValueError):
            tx.run(Query("RETURN 1", timeout=0.25))


def test_transaction_metadata(session):
    metadata_in = {"foo": "bar"}
    with session.begin_transaction(metadata=metadata_in) as tx:
        try:
            metadata_out = tx.run("CALL dbms.getTXMetaData").single().value()
        except ClientError as e:
            if e.code == "Neo.ClientError.Procedure.ProcedureNotFound":
                pytest.skip("Cannot assert correct metadata as Neo4j edition does not support procedure dbms.getTXMetaData")
            else:
                raise
        else:
            assert metadata_in == metadata_out


def test_transaction_timeout(driver):
    with driver.session() as s1:
        s1.run("CREATE (a:Node)").consume()
        with driver.session() as s2:
            tx1 = s1.begin_transaction()
            tx1.run("MATCH (a:Node) SET a.property = 1").consume()
            tx2 = s2.begin_transaction(timeout=0.25)
            try:
                tx2.run("MATCH (a:Node) SET a.property = 2").consume()
            # On 4.0 and older
            except TransientError:
                pass
            # On 4.1 and forward
            except ClientError:
                pass
            else:
                raise


# TODO: Re-enable and test when TC is available again
# def test_exit_after_explicit_close_should_be_silent(bolt_driver):
#     with bolt_driver.session() as s:
#         with s.begin_transaction() as tx:
#             assert not tx.closed()
#             tx.close()
#             assert tx.closed()
#         assert tx.closed()


@pytest.mark.skip(reason="This behaviour have changed. transaction.commit/rollback -> DISCARD n=-1, COMMIT/ROLL_BACK")
def test_should_sync_after_commit(session):
    tx = session.begin_transaction()
    result = tx.run("RETURN 1")
    tx.commit()
    buffer = result._records
    assert len(buffer) == 1
    assert buffer[0][0] == 1


@pytest.mark.skip(reason="This behaviour have changed. transaction.commit/rollback -> DISCARD n=-1, COMMIT/ROLL_BACK")
def test_should_sync_after_rollback(session):
    tx = session.begin_transaction()
    result = tx.run("RETURN 1")
    tx.rollback()
    buffer = result._records
    assert len(buffer) == 1
    assert buffer[0][0] == 1


def test_errors_on_run_transaction(session):
    tx = session.begin_transaction()
    with pytest.raises(TypeError):
        tx.run("CREATE (a:Thing {uuid:$uuid})", uuid=uuid4())
    tx.rollback()


def test_error_on_using_closed_transaction(session):
    tx = session.begin_transaction()
    tx.run("RETURN 1")
    tx.commit()
    with pytest.raises(TransactionError):
        tx.run("RETURN 1")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# tag::session-import[]
# end::session-import[]

# python -m pytest tests/integration/examples/test_session_example.py -s -v

def session_example(driver):
    with driver.session() as session:
        session.run("MATCH (_) DETACH DELETE _")

    # tag::session[]
    def add_person(name):
        with driver.session() as session:
            session.run("CREATE (a:Person {name: $name})", name=name)
    # end::session[]

    add_person("Alice")
    add_person("Bob")

    with driver.session() as session:
        persons = session.run("MATCH (a:Person) RETURN count(a)").single().value()

    with driver.session() as session:
        session.run("MATCH (_) DETACH DELETE _")

    return persons


def test_example(driver):
    persons = session_example(driver)
    assert persons == 2



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logging import getLogger
from ssl import SSLSocket

from neo4j._exceptions import (
    BoltError,
    BoltProtocolError,
)
from neo4j.api import (
    READ_ACCESS,
    Version,
)
from neo4j.exceptions import (
    ConfigurationError,
    DatabaseUnavailable,
    DriverError,
    ForbiddenOnReadOnlyDatabase,
    Neo4jError,
    NotALeader,
    ServiceUnavailable,
)
from neo4j.io import (
    Bolt,
    check_supported_server_product,
)
from neo4j.io._common import (
    CommitResponse,
    InitResponse,
    Response,
)


log = getLogger("neo4j")


class Bolt3(Bolt):
    """ Protocol handler for Bolt 3.

    This is supported by Neo4j versions 3.5, 4.0, 4.1 and 4.2.
    """

    PROTOCOL_VERSION = Version(3, 0)

    supports_multiple_results = False

    supports_multiple_databases = False

    @property
    def encrypted(self):
        return isinstance(self.socket, SSLSocket)

    @property
    def der_encoded_server_certificate(self):
        return self.socket.getpeercert(binary_form=True)

    @property
    def local_port(self):
        try:
            return self.socket.getsockname()[1]
        except OSError:
            return 0

    def get_base_headers(self):
        return {
            "user_agent": self.user_agent,
        }

    def hello(self):
        headers = self.get_base_headers()
        headers.update(self.auth_dict)
        logged_headers = dict(headers)
        if "credentials" in logged_headers:
            logged_headers["credentials"] = "*******"
        log.debug("[#%04X]  C: HELLO %r", self.local_port, logged_headers)
        self._append(b"\x01", (headers,),
                     response=InitResponse(self, on_success=self.server_info.update))
        self.send_all()
        self.fetch_all()
        check_supported_server_product(self.server_info.agent)

    def route(self, database=None, bookmarks=None):
        if database is not None:  # default database
            raise ConfigurationError("Database name parameter for selecting database is not "
                                     "supported in Bolt Protocol {!r}. Database name {!r}. "
                                     "Server Agent {!r}.".format(Bolt3.PROTOCOL_VERSION, database,
                                                                 self.server_info.agent))

        metadata = {}
        records = []

        def fail(md):
            from neo4j._exceptions import BoltRoutingError
            if md.get("code") == "Neo.ClientError.Procedure.ProcedureNotFound":
                raise BoltRoutingError("Server does not support routing", self.unresolved_address)
            else:
                raise BoltRoutingError("Routing support broken on server", self.unresolved_address)

        # Ignoring database and bookmarks because there is no multi-db support.
        # The bookmarks are only relevant for making sure a previously created
        # db exists before querying a routing table for it.
        self.run(
            "CALL dbms.cluster.routing.getRoutingTable($context)",  # This is an internal procedure call. Only available if the Neo4j 3.5 is setup with clustering.
            {"context": self.routing_context},
            mode="r",                                               # Bolt Protocol Version(3, 0) supports mode="r"
            on_success=metadata.update, on_failure=fail
        )
        self.pull(on_success=metadata.update, on_records=records.extend)
        self.send_all()
        self.fetch_all()
        routing_info = [dict(zip(metadata.get("fields", ()), values)) for values in records]
        return routing_info

    def run(self, query, parameters=None, mode=None, bookmarks=None, metadata=None, timeout=None, db=None, **handlers):
        if db is not None:
            raise ConfigurationError("Database name parameter for selecting database is not supported in Bolt Protocol {!r}. Database name {!r}.".format(Bolt3.PROTOCOL_VERSION, db))
        if not parameters:
            parameters = {}
        extra = {}
        if mode in (READ_ACCESS, "r"):
            extra["mode"] = "r"  # It will default to mode "w" if nothing is specified
        if bookmarks:
            try:
                extra["bookmarks"] = list(bookmarks)
            except TypeError:
                raise TypeError("Bookmarks must be provided within an iterable")
        if metadata:
            try:
                extra["tx_metadata"] = dict(metadata)
            except TypeError:
                raise TypeError("Metadata must be coercible to a dict")
        if timeout:
            try:
                extra["tx_timeout"] = int(1000 * timeout)
            except TypeError:
                raise TypeError("Timeout must be specified as a number of seconds")
        fields = (query, parameters, extra)
        log.debug("[#%04X]  C: RUN %s", self.local_port, " ".join(map(repr, fields)))
        if query.upper() == u"COMMIT":
            self._append(b"\x10", fields, CommitResponse(self, **handlers))
        else:
            self._append(b"\x10", fields, Response(self, **handlers))
        self._is_reset = False

    def discard(self, n=-1, qid=-1, **handlers):
        # Just ignore n and qid, it is not supported in the Bolt 3 Protocol.
        log.debug("[#%04X]  C: DISCARD_ALL", self.local_port)
        self._append(b"\x2F", (), Response(self, **handlers))

    def pull(self, n=-1, qid=-1, **handlers):
        # Just ignore n and qid, it is not supported in the Bolt 3 Protocol.
        log.debug("[#%04X]  C: PULL_ALL", self.local_port)
        self._append(b"\x3F", (), Response(self, **handlers))
        self._is_reset = False

    def begin(self, mode=None, bookmarks=None, metadata=None, timeout=None, db=None, **handlers):
        if db is not None:
            raise ConfigurationError("Database name parameter for selecting database is not supported in Bolt Protocol {!r}. Database name {!r}.".format(Bolt3.PROTOCOL_VERSION, db))
        extra = {}
        if mode in (READ_ACCESS, "r"):
            extra["mode"] = "r"  # It will default to mode "w" if nothing is specified
        if bookmarks:
            try:
                extra["bookmarks"] = list(bookmarks)
            except TypeError:
                raise TypeError("Bookmarks must be provided within an iterable")
        if metadata:
            try:
                extra["tx_metadata"] = dict(metadata)
            except TypeError:
                raise TypeError("Metadata must be coercible to a dict")
        if timeout:
            try:
                extra["tx_timeout"] = int(1000 * timeout)
            except TypeError:
                raise TypeError("Timeout must be specified as a number of seconds")
        log.debug("[#%04X]  C: BEGIN %r", self.local_port, extra)
        self._append(b"\x11", (extra,), Response(self, **handlers))
        self._is_reset = False

    def commit(self, **handlers):
        log.debug("[#%04X]  C: COMMIT", self.local_port)
        self._append(b"\x12", (), CommitResponse(self, **handlers))

    def rollback(self, **handlers):
        log.debug("[#%04X]  C: ROLLBACK", self.local_port)
        self._append(b"\x13", (), Response(self, **handlers))

    def reset(self):
        """ Add a RESET message to the outgoing queue, send
        it and consume all remaining messages.
        """

        def fail(metadata):
            raise BoltProtocolError("RESET failed %r" % metadata, address=self.unresolved_address)

        log.debug("[#%04X]  C: RESET", self.local_port)
        self._append(b"\x0F", response=Response(self, on_failure=fail))
        self.send_all()
        self.fetch_all()
        self._is_reset = True

    def fetch_message(self):
        """ Receive at most one message from the server, if available.

        :return: 2-tuple of number of detail messages and number of summary
                 messages fetched
        """
        if self._closed:
            raise ServiceUnavailable("Failed to read from closed connection {!r} ({!r})".format(
                self.unresolved_address, self.server_info.address))

        if self._defunct:
            raise ServiceUnavailable("Failed to read from defunct connection {!r} ({!r})".format(
                self.unresolved_address, self.server_info.address))

        if not self.responses:
            return 0, 0

        # Receive exactly one message
        details, summary_signature, summary_metadata = next(self.inbox)

        if details:
            log.debug("[#%04X]  S: RECORD * %d", self.local_port, len(details))  # Do not log any data
            self.responses[0].on_records(details)

        if summary_signature is None:
            return len(details), 0

        response = self.responses.popleft()
        response.complete = True
        if summary_signature == b"\x70":
            log.debug("[#%04X]  S: SUCCESS %r", self.local_port, summary_metadata)
            response.on_success(summary_metadata or {})
        elif summary_signature == b"\x7E":
            log.debug("[#%04X]  S: IGNORED", self.local_port)
            response.on_ignored(summary_metadata or {})
        elif summary_signature == b"\x7F":
            log.debug("[#%04X]  S: FAILURE %r", self.local_port, summary_metadata)
            try:
                response.on_failure(summary_metadata or {})
            except (ServiceUnavailable, DatabaseUnavailable):
                if self.pool:
                    self.pool.deactivate(address=self.unresolved_address),
                raise
            except (NotALeader, ForbiddenOnReadOnlyDatabase):
                if self.pool:
                    self.pool.on_write_failure(address=self.unresolved_address),
                raise
            except Neo4jError as e:
                if self.pool and e.invalidates_all_connections():
                    self.pool.mark_all_stale()
                raise
        else:
            raise BoltProtocolError("Unexpected response message with signature %02X" % summary_signature, address=self.unresolved_address)

        return len(details), 1

    def close(self):
        """ Close the connection.
        """
        if not self._closed:
            if not self._defunct:
                log.debug("[#%04X]  C: GOODBYE", self.local_port)
                self._append(b"\x02", ())
                try:
                    self._send_all()
                except (OSError, BoltError, DriverError):
                    pass
            log.debug("[#%04X]  C: <CLOSE>", self.local_port)
            try:
                self.socket.close()
            except OSError:
                pass
            finally:
                self._closed = True

    def closed(self):
        return self._closed

    def defunct(self):
        return self._defunct



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest import (
    mock,
    TestCase,
)
import pytest
from threading import (
    Condition,
    Event,
    Lock,
    Thread,
)
import time

from neo4j import (
    Config,
    PoolConfig,
    WorkspaceConfig,
)
from neo4j.io import (
    Bolt,
    BoltPool,
    IOPool
)
from neo4j.exceptions import (
    ClientError,
    ServiceUnavailable,
)


class FakeSocket:
    def __init__(self, address):
        self.address = address

    def getpeername(self):
        return self.address

    def sendall(self, data):
        return

    def close(self):
        return


class QuickConnection:

    def __init__(self, socket):
        self.socket = socket
        self.address = socket.getpeername()

    @property
    def is_reset(self):
        return True

    def stale(self):
        return False

    def reset(self):
        pass

    def close(self):
        self.socket.close()

    def closed(self):
        return False

    def defunct(self):
        return False

    def timedout(self):
        return False


class FakeBoltPool(IOPool):

    def __init__(self, address, *, auth=None, **config):
        self.pool_config, self.workspace_config = Config.consume_chain(config, PoolConfig, WorkspaceConfig)
        if config:
            raise ValueError("Unexpected config keys: %s" % ", ".join(config.keys()))

        def opener(addr, timeout):
            return QuickConnection(FakeSocket(addr))

        super().__init__(opener, self.pool_config, self.workspace_config)
        self.address = address

    def acquire(self, access_mode=None, timeout=None, database=None):
        return self._acquire(self.address, timeout)


class BoltTestCase(TestCase):

    def test_open(self):
        with pytest.raises(ServiceUnavailable):
            connection = Bolt.open(("localhost", 9999), auth=("test", "test"))

    def test_open_timeout(self):
        with pytest.raises(ServiceUnavailable):
            connection = Bolt.open(("localhost", 9999), auth=("test", "test"), timeout=1)

    def test_ping(self):
        protocol_version = Bolt.ping(("localhost", 9999))
        assert protocol_version is None

    def test_ping_timeout(self):
        protocol_version = Bolt.ping(("localhost", 9999), timeout=1)
        assert protocol_version is None


class MultiEvent:
    # Adopted from threading.Event

    def __init__(self):
        super().__init__()
        self._cond = Condition(Lock())
        self._counter = 0

    def _reset_internal_locks(self):
        # private!  called by Thread._reset_internal_locks by _after_fork()
        self._cond.__init__(Lock())

    def counter(self):
        return self._counter

    def increment(self):
        with self._cond:
            self._counter += 1
            self._cond.notify_all()

    def decrement(self):
        with self._cond:
            self._counter -= 1
            self._cond.notify_all()

    def clear(self):
        with self._cond:
            self._counter = 0
            self._cond.notify_all()

    def wait(self, value=0, timeout=None):
        with self._cond:
            t_start = time.time()
            while True:
                if value == self._counter:
                    return True
                if timeout is None:
                    time_left = None
                else:
                    time_left = timeout - (time.time() - t_start)
                    if time_left <= 0:
                        return False
                if not self._cond.wait(time_left):
                    return False


class ConnectionPoolTestCase(TestCase):

    def setUp(self):
        self.pool = FakeBoltPool(("127.0.0.1", 7687))

    def tearDown(self):
        self.pool.close()

    def assert_pool_size(self, address, expected_active, expected_inactive, pool=None):
        if pool is None:
            pool = self.pool
        try:
            connections = pool.connections[address]
        except KeyError:
            self.assertEqual(0, expected_active)
            self.assertEqual(0, expected_inactive)
        else:
            self.assertEqual(expected_active, len([cx for cx in connections if cx.in_use]))
            self.assertEqual(expected_inactive, len([cx for cx in connections if not cx.in_use]))

    def test_can_acquire(self):
        address = ("127.0.0.1", 7687)
        connection = self.pool._acquire(address, timeout=3)
        assert connection.address == address
        self.assert_pool_size(address, 1, 0)

    def test_can_acquire_twice(self):
        address = ("127.0.0.1", 7687)
        connection_1 = self.pool._acquire(address, timeout=3)
        connection_2 = self.pool._acquire(address, timeout=3)
        assert connection_1.address == address
        assert connection_2.address == address
        assert connection_1 is not connection_2
        self.assert_pool_size(address, 2, 0)

    def test_can_acquire_two_addresses(self):
        address_1 = ("127.0.0.1", 7687)
        address_2 = ("127.0.0.1", 7474)
        connection_1 = self.pool._acquire(address_1, timeout=3)
        connection_2 = self.pool._acquire(address_2, timeout=3)
        assert connection_1.address == address_1
        assert connection_2.address == address_2
        self.assert_pool_size(address_1, 1, 0)
        self.assert_pool_size(address_2, 1, 0)

    def test_can_acquire_and_release(self):
        address = ("127.0.0.1", 7687)
        connection = self.pool._acquire(address, timeout=3)
        self.assert_pool_size(address, 1, 0)
        self.pool.release(connection)
        self.assert_pool_size(address, 0, 1)

    def test_releasing_twice(self):
        address = ("127.0.0.1", 7687)
        connection = self.pool._acquire(address, timeout=3)
        self.pool.release(connection)
        self.assert_pool_size(address, 0, 1)
        self.pool.release(connection)
        self.assert_pool_size(address, 0, 1)

    def test_in_use_count(self):
        address = ("127.0.0.1", 7687)
        self.assertEqual(self.pool.in_use_connection_count(address), 0)
        connection = self.pool._acquire(address, timeout=3)
        self.assertEqual(self.pool.in_use_connection_count(address), 1)
        self.pool.release(connection)
        self.assertEqual(self.pool.in_use_connection_count(address), 0)

    def test_max_conn_pool_size(self):
        with FakeBoltPool((), max_connection_pool_size=1) as pool:
            address = ("127.0.0.1", 7687)
            pool._acquire(address, timeout=0)
            self.assertEqual(pool.in_use_connection_count(address), 1)
            with self.assertRaises(ClientError):
                pool._acquire(address, timeout=0)
            self.assertEqual(pool.in_use_connection_count(address), 1)

    def test_multithread(self):
        with FakeBoltPool((), max_connection_pool_size=5) as pool:
            address = ("127.0.0.1", 7687)
            acquired_counter = MultiEvent()
            release_event = Event()

            # start 10 threads competing for connections from a pool of size 5
            threads = []
            for i in range(10):
                t = Thread(
                    target=acquire_release_conn,
                    args=(pool, address, acquired_counter, release_event),
                    daemon=True
                )
                t.start()
                threads.append(t)

            if not acquired_counter.wait(5, timeout=1):
                raise RuntimeError("Acquire threads not fast enough")
            # The pool size should be 5, all are in-use
            self.assert_pool_size(address, 5, 0, pool)
            # Now we allow thread to release connections they obtained from pool
            release_event.set()

            # wait for all threads to release connections back to pool
            for t in threads:
                t.join(timeout=1)
            # The pool size is still 5, but all are free
            self.assert_pool_size(address, 0, 5, pool)

    def test_reset_when_released(self):
        def test(is_reset):
            with mock.patch(__name__ + ".QuickConnection.is_reset",
                            new_callable=mock.PropertyMock) as is_reset_mock:
                with mock.patch(__name__ + ".QuickConnection.reset",
                                new_callable=mock.MagicMock) as reset_mock:
                    is_reset_mock.return_value = is_reset
                    connection = self.pool._acquire(address, timeout=3)
                    self.assertIsInstance(connection, QuickConnection)
                    self.assertEqual(is_reset_mock.call_count, 0)
                    self.assertEqual(reset_mock.call_count, 0)
                    self.pool.release(connection)
                    self.assertEqual(is_reset_mock.call_count, 1)
                    self.assertEqual(reset_mock.call_count, int(not is_reset))

        address = ("127.0.0.1", 7687)
        for is_reset in (True, False):
            with self.subTest():
                test(is_reset)


def acquire_release_conn(pool, address, acquired_counter, release_event):
    conn = pool._acquire(address, timeout=3)
    acquired_counter.increment()
    release_event.wait()
    pool.release(conn)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# tag::result-retain-import[]
# end::result-retain-import[]


# python -m pytest tests/integration/examples/test_result_retain_example.py -s -v

def result_retain_example(driver):
    with driver.session() as session:
        session.run("MATCH (_) DETACH DELETE _").consume()

    with driver.session() as session:
        session.run("CREATE (a:Person {name: $name}) RETURN a", name="Alice").single().value()
        session.run("CREATE (a:Person {name: $name}) RETURN a", name="Bob").single().value()

    # tag::result-retain[]
    def add_employee_to_company(tx, person, company_name):
        tx.run("MATCH (emp:Person {name: $person_name}) "
               "MERGE (com:Company {name: $company_name}) "
               "MERGE (emp)-[:WORKS_FOR]->(com)",
               person_name=person["name"], company_name=company_name)
        return 1

    def match_person_nodes(tx):
        return list(tx.run("MATCH (a:Person) RETURN a.name AS name"))

    def add_employees(company_name):
        employees = 0
        with driver.session() as session:
            persons = session.read_transaction(match_person_nodes)

            for person in persons:
                employees += session.write_transaction(add_employee_to_company, person, company_name)

        return employees
    # end::result-retain[]

    employees = add_employees(company_name="Neo4j")

    def count_employees(company_name):
        with driver.session() as session:
            head_count = session.run(
                "MATCH (emp:Person)-[:WORKS_FOR]->(com:Company) "
                "WHERE com.name = $company_name "
                "RETURN count(emp)",
                company_name=company_name).single().value()

        return head_count

    head_count = count_employees(company_name="Neo4j")

    with driver.session() as session:
        session.run("MATCH (_) DETACH DELETE _").consume()

    return employees, head_count


def test_example(driver):
    employees, head_count = result_retain_example(driver)
    assert employees == 2
    assert head_count == 2



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from io import BytesIO
from struct import pack as struct_pack, unpack as struct_unpack

import pytest

from neo4j.io._common import MessageInbox
from neo4j.packstream import Packer, UnpackableBuffer, Unpacker


class FakeSocket:

    def __init__(self, address):
        self.address = address
        self.captured = b""
        self.messages = MessageInbox(self, on_error=print)

    def getsockname(self):
        return "127.0.0.1", 0xFFFF

    def getpeername(self):
        return self.address

    def recv_into(self, buffer, nbytes):
        data = self.captured[:nbytes]
        actual = len(data)
        buffer[:actual] = data
        self.captured = self.captured[actual:]
        return actual

    def sendall(self, data):
        self.captured += data

    def close(self):
        return

    def pop_message(self):
        return self.messages.pop()


class FakeSocket2:

    def __init__(self, address=None, on_send=None):
        self.address = address
        self.recv_buffer = bytearray()
        self._messages = MessageInbox(self, on_error=print)
        self.on_send = on_send

    def getsockname(self):
        return "127.0.0.1", 0xFFFF

    def getpeername(self):
        return self.address

    def recv_into(self, buffer, nbytes):
        data = self.recv_buffer[:nbytes]
        actual = len(data)
        buffer[:actual] = data
        self.recv_buffer = self.recv_buffer[actual:]
        return actual

    def sendall(self, data):
        if callable(self.on_send):
            self.on_send(data)

    def close(self):
        return

    def inject(self, data):
        self.recv_buffer += data

    def pop_chunk(self):
        chunk_size, = struct_unpack(">H", self.recv_buffer[:2])
        print("CHUNK SIZE %r" % chunk_size)
        end = 2 + chunk_size
        chunk_data, self.recv_buffer = self.recv_buffer[2:end], self.recv_buffer[end:]
        return chunk_data

    def pop_message(self):
        data = bytearray()
        while True:
            chunk = self.pop_chunk()
            print("CHUNK %r" % chunk)
            if chunk:
                data.extend(chunk)
            elif data:
                break       # end of message
            else:
                continue    # NOOP
        header = data[0]
        n_fields = header % 0x10
        tag = data[1]
        buffer = UnpackableBuffer(data[2:])
        unpacker = Unpacker(buffer)
        fields = [unpacker.unpack() for _ in range(n_fields)]
        return tag, fields

    def send_message(self, tag, *fields):
        data = self.encode_message(tag, *fields)
        self.sendall(struct_pack(">H", len(data)) + data + b"\x00\x00")

    @classmethod
    def encode_message(cls, tag, *fields):
        b = BytesIO()
        packer = Packer(b)
        for field in fields:
            packer.pack(field)
        return bytearray([0xB0 + len(fields), tag]) + b.getvalue()


class FakeSocketPair:

    def __init__(self, address):
        self.client = FakeSocket2(address)
        self.server = FakeSocket2()
        self.client.on_send = self.server.inject
        self.server.on_send = self.client.inject


@pytest.fixture
def fake_socket():
    return FakeSocket


@pytest.fixture
def fake_socket_2():
    return FakeSocket2


@pytest.fixture
def fake_socket_pair():
    return FakeSocketPair



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from os import path

import neo4j
import testkitbackend.fromtestkit as fromtestkit
import testkitbackend.totestkit as totestkit
from testkitbackend.fromtestkit import to_meta_and_timeout


class FrontendError(Exception):
    pass


def load_config():
    with open(path.join(path.dirname(__file__), "test_config.json"), "r") as fd:
        config = json.load(fd)
    return (config["skips"],
            [k for k, v in config["features"].items() if v is True])


SKIPPED_TESTS, FEATURES = load_config()


def StartTest(backend, data):
    if data["testName"] in SKIPPED_TESTS:
        backend.send_response("SkipTest",
                              {"reason": SKIPPED_TESTS[data["testName"]]})
    else:
        backend.send_response("RunTest", {})


def GetFeatures(backend, data):
    backend.send_response("FeatureList", {"features": FEATURES})


def NewDriver(backend, data):
    auth_token = data["authorizationToken"]["data"]
    data["authorizationToken"].mark_item_as_read_if_equals(
        "name", "AuthorizationToken"
    )
    auth = neo4j.Auth(
            auth_token["scheme"], auth_token["principal"],
            auth_token["credentials"], realm=auth_token["realm"])
    auth_token.mark_item_as_read_if_equals("ticket", "")
    resolver = None
    if data["resolverRegistered"] or data["domainNameResolverRegistered"]:
        resolver = resolution_func(backend, data["resolverRegistered"],
                                   data["domainNameResolverRegistered"])
    connection_timeout = data.get("connectionTimeoutMs", None)
    if connection_timeout is not None:
        connection_timeout /= 1000
    data.mark_item_as_read("domainNameResolverRegistered")
    driver = neo4j.GraphDatabase.driver(
        data["uri"], auth=auth, user_agent=data["userAgent"],
        resolver=resolver, connection_timeout=connection_timeout
    )
    key = backend.next_key()
    backend.drivers[key] = driver
    backend.send_response("Driver", {"id": key})


def VerifyConnectivity(backend, data):
    driver_id = data["driverId"]
    driver = backend.drivers[driver_id]
    driver.verify_connectivity()
    backend.send_response("Driver", {"id": driver_id})


def CheckMultiDBSupport(backend, data):
    driver_id = data["driverId"]
    driver = backend.drivers[driver_id]
    backend.send_response(
        "MultiDBSupport",
        {"id": backend.next_key(), "available": driver.supports_multi_db()}
    )


def resolution_func(backend, custom_resolver=False, custom_dns_resolver=False):
    # This solution (putting custom resolution together with DNS resolution into
    # one function only works because the Python driver calls the custom
    # resolver function for every connection, which is not true for all drivers.
    # Properly exposing a way to change the DNS lookup behavior is not possible
    # without changing the driver's code.
    assert custom_resolver or custom_dns_resolver

    def resolve(address):
        addresses = [":".join(map(str, address))]
        if custom_resolver:
            key = backend.next_key()
            backend.send_response("ResolverResolutionRequired", {
                "id": key,
                "address": addresses[0]
            })
            if not backend.process_request():
                # connection was closed before end of next message
                return []
            if key not in backend.custom_resolutions:
                raise RuntimeError(
                    "Backend did not receive expected "
                    "ResolverResolutionCompleted message for id %s" % key
                )
            addresses = backend.custom_resolutions.pop(key)
        if custom_dns_resolver:
            dns_resolved_addresses = []
            for address in addresses:
                key = backend.next_key()
                address = address.rsplit(":", 1)
                backend.send_response("DomainNameResolutionRequired", {
                    "id": key,
                    "name": address[0]
                })
                if not backend.process_request():
                    # connection was closed before end of next message
                    return []
                if key not in backend.dns_resolutions:
                    raise RuntimeError(
                        "Backend did not receive expected "
                        "DomainNameResolutionCompleted message for id %s" % key
                    )
                dns_resolved_addresses += list(map(
                    lambda a: ":".join((a, *address[1:])),
                    backend.dns_resolutions.pop(key)
                ))

            addresses = dns_resolved_addresses

        return list(map(neo4j.Address.parse, addresses))

    return resolve


def ResolverResolutionCompleted(backend, data):
    backend.custom_resolutions[data["requestId"]] = data["addresses"]


def DomainNameResolutionCompleted(backend, data):
    backend.dns_resolutions[data["requestId"]] = data["addresses"]


def DriverClose(backend, data):
    key = data["driverId"]
    driver = backend.drivers[key]
    driver.close()
    backend.send_response("Driver", {"id": key})


class SessionTracker:
    """ Keeps some extra state about the tracked session
    """

    def __init__(self, session):
        self.session = session
        self.state = ""
        self.error_id = ""


def NewSession(backend, data):
    driver = backend.drivers[data["driverId"]]
    access_mode = data["accessMode"]
    if access_mode == "r":
        access_mode = neo4j.READ_ACCESS
    elif access_mode == "w":
        access_mode = neo4j.WRITE_ACCESS
    else:
        raise Exception("Unknown access mode:" + access_mode)
    config = {
            "default_access_mode": access_mode,
            "bookmarks": data["bookmarks"],
            "database": data["database"],
            "fetch_size": data.get("fetchSize", None)
    }
    session = driver.session(**config)
    key = backend.next_key()
    backend.sessions[key] = SessionTracker(session)
    backend.send_response("Session", {"id": key})


def SessionRun(backend, data):
    session = backend.sessions[data["sessionId"]].session
    query, params = fromtestkit.to_query_and_params(data)
    result = session.run(query, parameters=params)
    key = backend.next_key()
    backend.results[key] = result
    backend.send_response("Result", {"id": key, "keys": result.keys()})


def SessionClose(backend, data):
    key = data["sessionId"]
    session = backend.sessions[key].session
    session.close()
    del backend.sessions[key]
    backend.send_response("Session", {"id": key})


def SessionBeginTransaction(backend, data):
    key = data["sessionId"]
    session = backend.sessions[key].session
    metadata, timeout = to_meta_and_timeout(data)
    tx = session.begin_transaction(metadata=metadata, timeout=timeout)
    key = backend.next_key()
    backend.transactions[key] = tx
    backend.send_response("Transaction", {"id": key})


def SessionReadTransaction(backend, data):
    transactionFunc(backend, data, True)


def SessionWriteTransaction(backend, data):
    transactionFunc(backend, data, False)


def transactionFunc(backend, data, is_read):
    key = data["sessionId"]
    session_tracker = backend.sessions[key]
    session = session_tracker.session
    metadata, timeout = to_meta_and_timeout(data)

    @neo4j.unit_of_work(metadata=metadata, timeout=timeout)
    def func(tx):
        txkey = backend.next_key()
        backend.transactions[txkey] = tx
        session_tracker.state = ''
        backend.send_response("RetryableTry", {"id": txkey})

        cont = True
        while cont:
            cont = backend.process_request()
            if session_tracker.state == '+':
                cont = False
            elif session_tracker.state == '-':
                if session_tracker.error_id:
                    raise backend.errors[session_tracker.error_id]
                else:
                    raise FrontendError("Client said no")

    if is_read:
        session.read_transaction(func)
    else:
        session.write_transaction(func)
    backend.send_response("RetryableDone", {})


def SessionLastBookmarks(backend, data):
    key = data["sessionId"]
    session = backend.sessions[key].session
    bookmark = session.last_bookmark()
    bookmarks = []
    if bookmark:
        bookmarks.append(bookmark)
    backend.send_response("Bookmarks", {"bookmarks": bookmarks})


def TransactionRun(backend, data):
    key = data["txId"]
    tx = backend.transactions[key]
    cypher, params = fromtestkit.to_cypher_and_params(data)
    result = tx.run(cypher, parameters=params)
    key = backend.next_key()
    backend.results[key] = result
    backend.send_response("Result", {"id": key, "keys": result.keys()})


def TransactionCommit(backend, data):
    key = data["txId"]
    tx = backend.transactions[key]
    tx.commit()
    backend.send_response("Transaction", {"id": key})


def TransactionRollback(backend, data):
    key = data["txId"]
    tx = backend.transactions[key]
    tx.rollback()
    backend.send_response("Transaction", {"id": key})


def ResultNext(backend, data):
    result = backend.results[data["resultId"]]
    try:
        record = next(iter(result))
    except StopIteration:
        backend.send_response("NullRecord", {})
        return
    backend.send_response("Record", totestkit.record(record))


def ResultConsume(backend, data):
    result = backend.results[data["resultId"]]
    summary = result.consume()
    from neo4j.work.summary import ResultSummary
    assert isinstance(summary, ResultSummary)
    backend.send_response("Summary", {
        "serverInfo": {
            "address": ":".join(map(str, summary.server.address)),
            "agent": summary.server.agent,
            "protocolVersion":
                ".".join(map(str, summary.server.protocol_version)),
        },
        "counters": None if not summary.counters else {
            "constraintsAdded": summary.counters.constraints_added,
            "constraintsRemoved": summary.counters.constraints_removed,
            "containsSystemUpdates": summary.counters.contains_system_updates,
            "containsUpdates": summary.counters.contains_updates,
            "indexesAdded": summary.counters.indexes_added,
            "indexesRemoved": summary.counters.indexes_removed,
            "labelsAdded": summary.counters.labels_added,
            "labelsRemoved": summary.counters.labels_removed,
            "nodesCreated": summary.counters.nodes_created,
            "nodesDeleted": summary.counters.nodes_deleted,
            "propertiesSet": summary.counters.properties_set,
            "relationshipsCreated": summary.counters.relationships_created,
            "relationshipsDeleted": summary.counters.relationships_deleted,
            "systemUpdates": summary.counters.system_updates,
        },
        "database": summary.database,
        "notifications": summary.notifications,
        "plan": summary.plan,
        "profile": summary.profile,
        "query": {
            "text": summary.query,
            "parameters": {k: totestkit.field(v)
                           for k, v in summary.parameters.items()},
        },
        "queryType": summary.query_type,
        "resultAvailableAfter": summary.result_available_after,
        "resultConsumedAfter": summary.result_consumed_after,
    })


def RetryablePositive(backend, data):
    key = data["sessionId"]
    session_tracker = backend.sessions[key]
    session_tracker.state = '+'


def RetryableNegative(backend, data):
    key = data["sessionId"]
    session_tracker = backend.sessions[key]
    session_tracker.state = '-'
    session_tracker.error_id = data.get('errorId', '')


def ForcedRoutingTableUpdate(backend, data):
    driver_id = data["driverId"]
    driver = backend.drivers[driver_id]
    database = data["database"]
    bookmarks = data["bookmarks"]
    with driver._pool.refresh_lock:
        driver._pool.create_routing_table(database)
        driver._pool.update_routing_table(database=database,
                                          bookmarks=bookmarks)
    backend.send_response("Driver", {"id": driver_id})


def GetRoutingTable(backend, data):
    driver_id = data["driverId"]
    database = data["database"]
    driver = backend.drivers[driver_id]
    routing_table = driver._pool.routing_tables[database]
    response_data = {
        "database": routing_table.database,
        "ttl": routing_table.ttl,
    }
    for role in ("routers", "readers", "writers"):
        addresses = routing_table.__getattribute__(role)
        response_data[role] = list(map(str, addresses))
    backend.send_response("RoutingTable", response_data)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import deque
from warnings import warn

from neo4j.data import DataDehydrator
from neo4j.exceptions import (
    ServiceUnavailable,
    SessionExpired,
)
from neo4j.work.summary import ResultSummary


class _ConnectionErrorHandler:
    """
    Wrapper class for handling connection errors.

    The class will wrap each method to invoke a callback if the method raises
    SessionExpired or ServiceUnavailable.
    The error will be re-raised after the callback.
    """

    def __init__(self, connection, on_network_error):
        """
        :param connection the connection object to warp
        :type connection Bolt
        :param on_network_error the function to be called when a method of
            connection raises of of the caught errors. The callback takes the
            error as argument.
        :type on_network_error callable
        """
        self._connection = connection
        self._on_network_error = on_network_error

    def __getattr__(self, item):
        connection_attr = getattr(self._connection, item)
        if not callable(connection_attr):
            return connection_attr

        def outer(func):
            def inner(*args, **kwargs):
                try:
                    func(*args, **kwargs)
                finally:
                    if self._connection.defunct():
                        self._on_network_error()
            return inner

        return outer(connection_attr)


class Result:
    """A handler for the result of Cypher query execution. Instances
    of this class are typically constructed and returned by
    :meth:`.Session.run` and :meth:`.Transaction.run`.
    """

    def __init__(self, connection, hydrant, fetch_size, on_closed,
                 on_network_error):
        self._connection = _ConnectionErrorHandler(connection, on_network_error)
        self._hydrant = hydrant
        self._on_closed = on_closed
        self._metadata = None
        self._record_buffer = deque()
        self._summary = None
        self._bookmark = None
        self._qid = -1
        self._fetch_size = fetch_size

        # states
        self._discarding = False    # discard the remainder of records
        self._attached = False      # attached to a connection
        # there are still more response messages we wait for
        self._streaming = False
        # there ar more records available to pull from the server
        self._has_more = False
        # the result has been fully iterated or consumed
        self._closed = False

    def _tx_ready_run(self, query, parameters, **kwparameters):
        # BEGIN+RUN does not carry any extra on the RUN message.
        # BEGIN {extra}
        # RUN "query" {parameters} {extra}
        self._run(query, parameters, None, None, None, **kwparameters)

    def _run(self, query, parameters, db, access_mode, bookmarks, **kwparameters):
        query_text = str(query)  # Query or string object
        query_metadata = getattr(query, "metadata", None)
        query_timeout = getattr(query, "timeout", None)

        parameters = DataDehydrator.fix_parameters(dict(parameters or {}, **kwparameters))

        self._metadata = {
            "query": query_text,
            "parameters": parameters,
            "server": self._connection.server_info,
        }

        def on_attached(metadata):
            self._metadata.update(metadata)
            self._qid = metadata.get("qid", -1)  # For auto-commit there is no qid and Bolt 3 do not support qid
            self._keys = metadata.get("fields")
            self._attached = True

        def on_failed_attach(metadata):
            self._metadata.update(metadata)
            self._attached = False
            self._on_closed()

        self._connection.run(
            query_text,
            parameters=parameters,
            mode=access_mode,
            bookmarks=bookmarks,
            metadata=query_metadata,
            timeout=query_timeout,
            db=db,
            on_success=on_attached,
            on_failure=on_failed_attach,
        )
        self._pull()
        self._connection.send_all()
        self._attach()

    def _pull(self):
        def on_records(records):
            if not self._discarding:
                self._record_buffer.extend(self._hydrant.hydrate_records(self._keys, records))

        def on_summary():
            self._attached = False
            self._on_closed()

        def on_failure(metadata):
            self._attached = False
            self._on_closed()

        def on_success(summary_metadata):
            self._streaming = False
            has_more = summary_metadata.get("has_more")
            self._has_more = bool(has_more)
            if has_more:
                return
            self._metadata.update(summary_metadata)
            self._bookmark = summary_metadata.get("bookmark")

        self._connection.pull(
            n=self._fetch_size,
            qid=self._qid,
            on_records=on_records,
            on_success=on_success,
            on_failure=on_failure,
            on_summary=on_summary,
        )
        self._streaming = True

    def _discard(self):
        def on_summary():
            self._attached = False
            self._on_closed()

        def on_failure(metadata):
            self._metadata.update(metadata)
            self._attached = False
            self._on_closed()

        def on_success(summary_metadata):
            self._streaming = False
            has_more = summary_metadata.get("has_more")
            self._has_more = bool(has_more)
            if has_more:
                return
            self._discarding = False
            self._metadata.update(summary_metadata)
            self._bookmark = summary_metadata.get("bookmark")

        # This was the last page received, discard the rest
        self._connection.discard(
            n=-1,
            qid=self._qid,
            on_success=on_success,
            on_failure=on_failure,
            on_summary=on_summary,
        )
        self._streaming = True

    def __iter__(self):
        """Iterator returning Records.
        :returns: Record, it is an immutable ordered collection of key-value pairs.
        :rtype: :class:`neo4j.Record`
        """
        while self._record_buffer or self._attached:
            if self._record_buffer:
                yield self._record_buffer.popleft()
            elif self._streaming:
                self._connection.fetch_message()
            elif self._discarding:
                self._discard()
                self._connection.send_all()
            elif self._has_more:
                self._pull()
                self._connection.send_all()

        self._closed = True

    def _attach(self):
        """Sets the Result object in an attached state by fetching messages from the connection to the buffer.
        """
        if self._closed is False:
            while self._attached is False:
                self._connection.fetch_message()

    def _buffer_all(self):
        """Sets the Result object in an detached state by fetching all records from the connection to the buffer.
        """
        record_buffer = deque()
        for record in self:
            record_buffer.append(record)
        self._closed = False
        self._record_buffer = record_buffer

    def _obtain_summary(self):
        """Obtain the summary of this result, buffering any remaining records.

        :returns: The :class:`neo4j.ResultSummary` for this result
        """
        if self._summary is None:
            if self._metadata:
                self._summary = ResultSummary(**self._metadata)
            elif self._connection:
                self._summary = ResultSummary(server=self._connection.server_info)

        return self._summary

    def keys(self):
        """The keys for the records in this result.

        :returns: tuple of key names
        :rtype: tuple
        """
        return self._keys

    def consume(self):
        """Consume the remainder of this result and return a :class:`neo4j.ResultSummary`.

        Example::

            def create_node_tx(tx, name):
                result = tx.run("CREATE (n:ExampleNode { name: $name }) RETURN n", name=name)
                record = result.single()
                value = record.value()
                info = result.consume()
                return value, info

            with driver.session() as session:
                node_id, info = session.write_transaction(create_node_tx, "example")

        Example::

            def get_two_tx(tx):
                result = tx.run("UNWIND [1,2,3,4] AS x RETURN x")
                values = []
                for ix, record in enumerate(result):
                    if x > 1:
                        break
                    values.append(record.values())
                info = result.consume()  # discard the remaining records if there are any
                # use the info for logging etc.
                return values, info

            with driver.session() as session:
                values, info = session.read_transaction(get_two_tx)

        :returns: The :class:`neo4j.ResultSummary` for this result
        """
        if self._closed is False:
            self._discarding = True
            for _ in self:
                pass

        return self._obtain_summary()

    def single(self):
        """Obtain the next and only remaining record from this result if available else return None.
        Calling this method always exhausts the result.

        A warning is generated if more than one record is available but
        the first of these is still returned.

        :returns: the next :class:`neo4j.Record` or :const:`None` if none remain
        :warns: if more than one record is available
        """
        records = list(self)  # TODO: exhausts the result with self.consume if there are more records.
        size = len(records)
        if size == 0:
            return None
        if size != 1:
            warn("Expected a result with a single record, but this result contains %d" % size)
        return records[0]

    def peek(self):
        """Obtain the next record from this result without consuming it.
        This leaves the record in the buffer for further processing.

        :returns: the next :class:`.Record` or :const:`None` if none remain
        """
        if self._record_buffer:
            return self._record_buffer[0]
        if not self._attached:
            return None
        while self._attached:
            self._connection.fetch_message()
            if self._record_buffer:
                return self._record_buffer[0]

        return None

    def graph(self):
        """Return a :class:`neo4j.graph.Graph` instance containing all the graph objects
        in the result. After calling this method, the result becomes
        detached, buffering all remaining records.

        :returns: a result graph
        :rtype: :class:`neo4j.graph.Graph`
        """
        self._buffer_all()
        return self._hydrant.graph

    def value(self, key=0, default=None):
        """Helper function that return the remainder of the result as a list of values.

        See :class:`neo4j.Record.value`

        :param key: field to return for each remaining record. Obtain a single value from the record by index or key.
        :param default: default value, used if the index of key is unavailable
        :returns: list of individual values
        :rtype: list
        """
        return [record.value(key, default) for record in self]

    def values(self, *keys):
        """Helper function that return the remainder of the result as a list of values lists.

        See :class:`neo4j.Record.values`

        :param keys: fields to return for each remaining record. Optionally filtering to include only certain values by index or key.
        :returns: list of values lists
        :rtype: list
        """
        return [record.values(*keys) for record in self]

    def data(self, *keys):
        """Helper function that return the remainder of the result as a list of dictionaries.

        See :class:`neo4j.Record.data`

        :param keys: fields to return for each remaining record. Optionally filtering to include only certain values by index or key.
        :returns: list of dictionaries
        :rtype: list
        """
        return [record.data(*keys) for record in self]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j.data import DataHydrator
from neo4j.packstream import Structure

# python -m pytest -s -v tests/unit/test_data.py


def test_can_hydrate_node_structure():
    hydrant = DataHydrator()

    struct = Structure(b'N', 123, ["Person"], {"name": "Alice"})
    alice, = hydrant.hydrate([struct])

    assert alice.id == 123
    assert alice.labels == {"Person"}
    assert set(alice.keys()) == {"name"}
    assert alice.get("name") == "Alice"


def test_hydrating_unknown_structure_returns_same():
    hydrant = DataHydrator()

    struct = Structure(b'?', "foo")
    mystery, = hydrant.hydrate([struct])

    assert mystery == struct


def test_can_hydrate_in_list():
    hydrant = DataHydrator()

    struct = Structure(b'N', 123, ["Person"], {"name": "Alice"})
    alice_in_list, = hydrant.hydrate([[struct]])

    assert isinstance(alice_in_list, list)

    alice, = alice_in_list

    assert alice.id == 123
    assert alice.labels == {"Person"}
    assert set(alice.keys()) == {"name"}
    assert alice.get("name") == "Alice"


def test_can_hydrate_in_dict():
    hydrant = DataHydrator()

    struct = Structure(b'N', 123, ["Person"], {"name": "Alice"})
    alice_in_dict, = hydrant.hydrate([{"foo": struct}])

    assert isinstance(alice_in_dict, dict)

    alice = alice_in_dict["foo"]

    assert alice.id == 123
    assert alice.labels == {"Person"}
    assert set(alice.keys()) == {"name"}
    assert alice.get("name") == "Alice"



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from unittest import TestCase
from datetime import (
    datetime,
    timedelta,
)
from pytz import (
    timezone,
    FixedOffset,
)

from neo4j.time import (
    DateTime,
    MIN_YEAR,
    MAX_YEAR,
    Duration,
)
from neo4j.time.arithmetic import (
    nano_add,
    nano_div,
)
from neo4j.time.clock_implementations import (
    Clock,
    ClockTime,
)
from neo4j.time.hydration import (
    hydrate_date,
    dehydrate_date,
    hydrate_time,
    dehydrate_time,
    hydrate_datetime,
    dehydrate_datetime,
    hydrate_duration,
    dehydrate_duration,
    dehydrate_timedelta,
)

timezone_us_eastern = timezone("US/Eastern")
timezone_utc = timezone("UTC")


class FixedClock(Clock):

    @classmethod
    def available(cls):
        return True

    @classmethod
    def precision(cls):
        return 12

    @classmethod
    def local_offset(cls):
        return ClockTime()

    def utc_time(self):
        return ClockTime(45296, 789000000)


class DateTimeTestCase(TestCase):

    def test_zero(self):
        t = DateTime(0, 0, 0, 0, 0, 0)
        self.assertEqual(t.year, 0)
        self.assertEqual(t.month, 0)
        self.assertEqual(t.day, 0)
        self.assertEqual(t.hour, 0)
        self.assertEqual(t.minute, 0)
        self.assertEqual(t.second, 0)

    def test_non_zero_naive(self):
        t = DateTime(2018, 4, 26, 23, 0, 17.914390409)
        self.assertEqual(t.year, 2018)
        self.assertEqual(t.month, 4)
        self.assertEqual(t.day, 26)
        self.assertEqual(t.hour, 23)
        self.assertEqual(t.minute, 0)
        self.assertEqual(t.second, 17.914390409)

    def test_year_lower_bound(self):
        with self.assertRaises(ValueError):
            _ = DateTime(MIN_YEAR - 1, 1, 1, 0, 0, 0)

    def test_year_upper_bound(self):
        with self.assertRaises(ValueError):
            _ = DateTime(MAX_YEAR + 1, 1, 1, 0, 0, 0)

    def test_month_lower_bound(self):
        with self.assertRaises(ValueError):
            _ = DateTime(2000, 0, 1, 0, 0, 0)

    def test_month_upper_bound(self):
        with self.assertRaises(ValueError):
            _ = DateTime(2000, 13, 1, 0, 0, 0)

    def test_day_zero(self):
        with self.assertRaises(ValueError):
            _ = DateTime(2000, 1, 0, 0, 0, 0)

    def test_day_30_of_29_day_month(self):
        with self.assertRaises(ValueError):
            _ = DateTime(2000, 2, 30, 0, 0, 0)

    def test_day_32_of_31_day_month(self):
        with self.assertRaises(ValueError):
            _ = DateTime(2000, 3, 32, 0, 0, 0)

    def test_day_31_of_30_day_month(self):
        with self.assertRaises(ValueError):
            _ = DateTime(2000, 4, 31, 0, 0, 0)

    def test_day_29_of_28_day_month(self):
        with self.assertRaises(ValueError):
            _ = DateTime(1999, 2, 29, 0, 0, 0)

    def test_last_day_of_month(self):
        t = DateTime(2000, 1, -1, 0, 0, 0)
        self.assertEqual(t.year, 2000)
        self.assertEqual(t.month, 1)
        self.assertEqual(t.day, 31)

    def test_today(self):
        t = DateTime.today()
        self.assertEqual(t.year, 1970)
        self.assertEqual(t.month, 1)
        self.assertEqual(t.day, 1)
        self.assertEqual(t.hour, 12)
        self.assertEqual(t.minute, 34)
        self.assertEqual(t.second, 56.789)

    def test_now_without_tz(self):
        t = DateTime.now()
        self.assertEqual(t.year, 1970)
        self.assertEqual(t.month, 1)
        self.assertEqual(t.day, 1)
        self.assertEqual(t.hour, 12)
        self.assertEqual(t.minute, 34)
        self.assertEqual(t.second, 56.789)
        self.assertIsNone(t.tzinfo)

    def test_now_with_tz(self):
        t = DateTime.now(timezone_us_eastern)
        self.assertEqual(t.year, 1970)
        self.assertEqual(t.month, 1)
        self.assertEqual(t.day, 1)
        self.assertEqual(t.hour, 7)
        self.assertEqual(t.minute, 34)
        self.assertEqual(t.second, 56.789)
        self.assertEqual(t.utcoffset(), timedelta(seconds=-18000))
        self.assertEqual(t.dst(), timedelta())
        self.assertEqual(t.tzname(), "EST")

    def test_utc_now(self):
        t = DateTime.utc_now()
        self.assertEqual(t.year, 1970)
        self.assertEqual(t.month, 1)
        self.assertEqual(t.day, 1)
        self.assertEqual(t.hour, 12)
        self.assertEqual(t.minute, 34)
        self.assertEqual(t.second, 56.789)
        self.assertIsNone(t.tzinfo)

    def test_from_timestamp(self):
        t = DateTime.from_timestamp(0)
        self.assertEqual(t.year, 1970)
        self.assertEqual(t.month, 1)
        self.assertEqual(t.day, 1)
        self.assertEqual(t.hour, 0)
        self.assertEqual(t.minute, 0)
        self.assertEqual(t.second, 0.0)
        self.assertIsNone(t.tzinfo)

    def test_from_overflowing_timestamp(self):
        with self.assertRaises(ValueError):
            _ = DateTime.from_timestamp(999999999999999999)

    def test_from_timestamp_with_tz(self):
        t = DateTime.from_timestamp(0, timezone_us_eastern)
        self.assertEqual(t.year, 1969)
        self.assertEqual(t.month, 12)
        self.assertEqual(t.day, 31)
        self.assertEqual(t.hour, 19)
        self.assertEqual(t.minute, 0)
        self.assertEqual(t.second, 0.0)
        self.assertEqual(t.utcoffset(), timedelta(seconds=-18000))
        self.assertEqual(t.dst(), timedelta())
        self.assertEqual(t.tzname(), "EST")

    def test_conversion_to_t(self):
        dt = DateTime(2018, 4, 26, 23, 0, 17.914390409)
        t = dt.to_clock_time()
        self.assertEqual(t, ClockTime(63660380417, 914390409))

    def test_add_timedelta(self):
        dt1 = DateTime(2018, 4, 26, 23, 0, 17.914390409)
        delta = timedelta(days=1)
        dt2 = dt1 + delta
        self.assertEqual(dt2, DateTime(2018, 4, 27, 23, 0, 17.914390409))

    def test_subtract_datetime_1(self):
        dt1 = DateTime(2018, 4, 26, 23, 0, 17.914390409)
        dt2 = DateTime(2018, 1, 1, 0, 0, 0.0)
        t = dt1 - dt2
        self.assertEqual(t, Duration(months=3, days=25, hours=23, seconds=17.914390409))

    def test_subtract_datetime_2(self):
        dt1 = DateTime(2018, 4, 1, 23, 0, 17.914390409)
        dt2 = DateTime(2018, 1, 26, 0, 0, 0.0)
        t = dt1 - dt2
        self.assertEqual(t, Duration(months=3, days=-25, hours=23, seconds=17.914390409))

    def test_subtract_native_datetime_1(self):
        dt1 = DateTime(2018, 4, 26, 23, 0, 17.914390409)
        dt2 = datetime(2018, 1, 1, 0, 0, 0)
        t = dt1 - dt2
        self.assertEqual(t, timedelta(days=115, hours=23, seconds=17.914390409))

    def test_subtract_native_datetime_2(self):
        dt1 = DateTime(2018, 4, 1, 23, 0, 17.914390409)
        dt2 = datetime(2018, 1, 26, 0, 0, 0)
        t = dt1 - dt2
        self.assertEqual(t, timedelta(days=65, hours=23, seconds=17.914390409))

    def test_normalization(self):
        ndt1 = timezone_us_eastern.normalize(DateTime(2018, 4, 27, 23, 0, 17, tzinfo=timezone_us_eastern))
        ndt2 = timezone_us_eastern.normalize(datetime(2018, 4, 27, 23, 0, 17, tzinfo=timezone_us_eastern))
        self.assertEqual(ndt1, ndt2)

    def test_localization(self):
        ldt1 = timezone_us_eastern.localize(datetime(2018, 4, 27, 23, 0, 17))
        ldt2 = timezone_us_eastern.localize(DateTime(2018, 4, 27, 23, 0, 17))
        self.assertEqual(ldt1, ldt2)

    def test_from_native(self):
        native = datetime(2018, 10, 1, 12, 34, 56, 789123)
        dt = DateTime.from_native(native)
        self.assertEqual(dt.year, native.year)
        self.assertEqual(dt.month, native.month)
        self.assertEqual(dt.day, native.day)
        self.assertEqual(dt.hour, native.hour)
        self.assertEqual(dt.minute, native.minute)
        self.assertEqual(dt.second, nano_add(native.second, nano_div(native.microsecond, 1000000)))

    def test_to_native(self):
        dt = DateTime(2018, 10, 1, 12, 34, 56.789123456)
        native = dt.to_native()
        self.assertEqual(dt.year, native.year)
        self.assertEqual(dt.month, native.month)
        self.assertEqual(dt.day, native.day)
        self.assertEqual(dt.hour, native.hour)
        self.assertEqual(dt.minute, native.minute)
        self.assertEqual(56.789123, nano_add(native.second, nano_div(native.microsecond, 1000000)))

    def test_iso_format(self):
        dt = DateTime(2018, 10, 1, 12, 34, 56.789123456)
        self.assertEqual("2018-10-01T12:34:56.789123456", dt.iso_format())

    def test_iso_format_with_trailing_zeroes(self):
        dt = DateTime(2018, 10, 1, 12, 34, 56.789)
        self.assertEqual("2018-10-01T12:34:56.789000000", dt.iso_format())

    def test_iso_format_with_tz(self):
        dt = timezone_us_eastern.localize(DateTime(2018, 10, 1, 12, 34, 56.789123456))
        self.assertEqual("2018-10-01T12:34:56.789123456-04:00", dt.iso_format())

    def test_iso_format_with_tz_and_trailing_zeroes(self):
        dt = timezone_us_eastern.localize(DateTime(2018, 10, 1, 12, 34, 56.789))
        self.assertEqual("2018-10-01T12:34:56.789000000-04:00", dt.iso_format())

    def test_from_iso_format_hour_only(self):
        expected = DateTime(2018, 10, 1, 12, 0, 0)
        actual = DateTime.from_iso_format("2018-10-01T12")
        self.assertEqual(expected, actual)

    def test_from_iso_format_hour_and_minute(self):
        expected = DateTime(2018, 10, 1, 12, 34, 0)
        actual = DateTime.from_iso_format("2018-10-01T12:34")
        self.assertEqual(expected, actual)

    def test_from_iso_format_hour_minute_second(self):
        expected = DateTime(2018, 10, 1, 12, 34, 56)
        actual = DateTime.from_iso_format("2018-10-01T12:34:56")
        self.assertEqual(expected, actual)

    def test_from_iso_format_hour_minute_second_milliseconds(self):
        expected = DateTime(2018, 10, 1, 12, 34, 56.123)
        actual = DateTime.from_iso_format("2018-10-01T12:34:56.123")
        self.assertEqual(expected, actual)

    def test_from_iso_format_hour_minute_second_microseconds(self):
        expected = DateTime(2018, 10, 1, 12, 34, 56.123456)
        actual = DateTime.from_iso_format("2018-10-01T12:34:56.123456")
        self.assertEqual(expected, actual)

    def test_from_iso_format_hour_minute_second_nanoseconds(self):
        expected = DateTime(2018, 10, 1, 12, 34, 56.123456789)
        actual = DateTime.from_iso_format("2018-10-01T12:34:56.123456789")
        self.assertEqual(expected, actual)

    def test_from_iso_format_with_positive_tz(self):
        expected = DateTime(2018, 10, 1, 12, 34, 56.123456789, tzinfo=FixedOffset(754))
        actual = DateTime.from_iso_format("2018-10-01T12:34:56.123456789+12:34")
        self.assertEqual(expected, actual)

    def test_from_iso_format_with_negative_tz(self):
        expected = DateTime(2018, 10, 1, 12, 34, 56.123456789, tzinfo=FixedOffset(-754))
        actual = DateTime.from_iso_format("2018-10-01T12:34:56.123456789-12:34")
        self.assertEqual(expected, actual)

    def test_from_iso_format_with_positive_long_tz(self):
        expected = DateTime(2018, 10, 1, 12, 34, 56.123456789, tzinfo=FixedOffset(754))
        actual = DateTime.from_iso_format("2018-10-01T12:34:56.123456789+12:34:56.123456")
        self.assertEqual(expected, actual)

    def test_from_iso_format_with_negative_long_tz(self):
        expected = DateTime(2018, 10, 1, 12, 34, 56.123456789, tzinfo=FixedOffset(-754))
        actual = DateTime.from_iso_format("2018-10-01T12:34:56.123456789-12:34:56.123456")
        self.assertEqual(expected, actual)

    def test_datetime_copy(self):
        d = DateTime(2010, 10, 1, 10, 0, 10)
        d2 = copy.copy(d)
        self.assertIsNot(d, d2)
        self.assertEqual(d, d2)

    def test_datetime_deep_copy(self):
        d = DateTime(2010, 10, 1, 10, 0, 12)
        d2 = copy.deepcopy(d)
        self.assertIsNot(d, d2)
        self.assertEqual(d, d2)


def test_iso_format_with_time_zone_case_1():
    # python -m pytest tests/unit/time/test_datetime.py -s -v -k test_iso_format_with_time_zone_case_1
    expected = DateTime(2019, 10, 30, 7, 54, 2.129790999, tzinfo=timezone_utc)
    assert expected.iso_format() == "2019-10-30T07:54:02.129790999+00:00"
    assert expected.tzinfo == FixedOffset(0)
    actual = DateTime.from_iso_format("2019-10-30T07:54:02.129790999+00:00")
    assert expected == actual


def test_iso_format_with_time_zone_case_2():
    # python -m pytest tests/unit/time/test_datetime.py -s -v -k test_iso_format_with_time_zone_case_2
    expected = DateTime.from_iso_format("2019-10-30T07:54:02.129790999+01:00")
    assert expected.tzinfo == FixedOffset(60)
    assert expected.iso_format() == "2019-10-30T07:54:02.129790999+01:00"


def test_to_native_case_1():
    # python -m pytest tests/unit/time/test_datetime.py -s -v -k test_to_native_case_1
    dt = DateTime.from_iso_format("2019-10-30T12:34:56.789123456")
    native = dt.to_native()
    assert native.hour == dt.hour
    assert native.minute == dt.minute
    assert nano_add(native.second, nano_div(native.microsecond, 1000000)) == 56.789123
    assert native.tzinfo is None
    assert native.isoformat() == "2019-10-30T12:34:56.789123"


def test_to_native_case_2():
    # python -m pytest tests/unit/time/test_datetime.py -s -v -k test_to_native_case_2
    dt = DateTime.from_iso_format("2019-10-30T12:34:56.789123456+00:00")
    native = dt.to_native()
    assert native.hour == dt.hour
    assert native.minute == dt.minute
    assert nano_add(native.second, nano_div(native.microsecond, 1000000)) == 56.789123
    assert native.tzinfo == FixedOffset(0)
    assert native.isoformat() == "2019-10-30T12:34:56.789123+00:00"


def test_from_native_case_1():
    # python -m pytest tests/unit/time/test_datetime.py -s -v -k test_from_native_case_1
    native = datetime(2018, 10, 1, 12, 34, 56, 789123)
    dt = DateTime.from_native(native)
    assert dt.year == native.year
    assert dt.month == native.month
    assert dt.day == native.day
    assert dt.hour == native.hour
    assert dt.minute == native.minute
    assert dt.second == nano_add(native.second, nano_div(native.microsecond, 1000000))
    assert dt.tzinfo is None


def test_from_native_case_2():
    # python -m pytest tests/unit/time/test_datetime.py -s -v -k test_from_native_case_2
    native = datetime(2018, 10, 1, 12, 34, 56, 789123, FixedOffset(0))
    dt = DateTime.from_native(native)
    assert dt.year == native.year
    assert dt.month == native.month
    assert dt.day == native.day
    assert dt.hour == native.hour
    assert dt.minute == native.minute
    assert dt.second == nano_add(native.second, nano_div(native.microsecond, 1000000))
    assert dt.tzinfo == FixedOffset(0)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from pytest import fixture

from neo4j import (
    GraphDatabase,
    BoltDriver,
    Version,
    READ_ACCESS,
    WRITE_ACCESS,
    ResultSummary,
    unit_of_work,
    Transaction,
    Result,
    ServerInfo,
)
from neo4j.exceptions import (
    ServiceUnavailable,
    AuthError,
    ConfigurationError,
    ClientError,
)
from neo4j._exceptions import BoltHandshakeError
from neo4j.io._bolt3 import Bolt3


# TODO: this test will stay until a uniform behavior for `.single()` across the
#       drivers has been specified and tests are created in testkit
def test_normal_use_case(bolt_driver):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_normal_use_case
    session = bolt_driver.session()
    value = session.run("RETURN 1").single().value()
    assert value == 1


# TODO: this test will stay until a uniform behavior for `.encrypted` across the
#       drivers has been specified and tests are created in testkit
def test_encrypted_set_to_false_by_default(bolt_driver):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_encrypted_set_to_false_by_default
    assert bolt_driver.encrypted is False


def test_bolt_driver_fetch_size_config_case_on_close_result_consume(bolt_uri, auth):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_bolt_driver_fetch_size_config_case_on_close_result_consume
    try:
        with GraphDatabase.driver(bolt_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                result = session.run("UNWIND [1,2,3,4] AS x RETURN x")
                # Check the expected result with logging manually
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_bolt_driver_fetch_size_config_case_normal(bolt_uri, auth):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_bolt_driver_fetch_size_config_case_normal
    try:
        with GraphDatabase.driver(bolt_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = []
                result = session.run("UNWIND [1,2,3,4] AS x RETURN x")
                for record in result:
                    expected.append(record["x"])

        assert expected == [1, 2, 3, 4]
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_bolt_driver_fetch_size_config_run_consume_run(bolt_uri, auth):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_bolt_driver_fetch_size_config_run_consume_run
    try:
        with GraphDatabase.driver(bolt_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = []
                result1 = session.run("UNWIND [1,2,3,4] AS x RETURN x")
                result1.consume()
                result2 = session.run("UNWIND [5,6,7,8] AS x RETURN x")

                for record in result2:
                    expected.append(record["x"])

                result_summary = result2.consume()
                assert isinstance(result_summary, ResultSummary)

        assert expected == [5, 6, 7, 8]
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_bolt_driver_fetch_size_config_run_run(bolt_uri, auth):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_bolt_driver_fetch_size_config_run_run
    try:
        with GraphDatabase.driver(bolt_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = []
                result1 = session.run("UNWIND [1,2,3,4] AS x RETURN x")
                result2 = session.run("UNWIND [5,6,7,8] AS x RETURN x")

                for record in result2:
                    expected.append(record["x"])

                result_summary = result2.consume()
                assert isinstance(result_summary, ResultSummary)

        assert expected == [5, 6, 7, 8]
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_bolt_driver_read_transaction_fetch_size_config_normal_case(bolt_uri, auth):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_bolt_driver_read_transaction_fetch_size_config_normal_case
    @unit_of_work(timeout=3, metadata={"foo": "bar"})
    def unwind(transaction):
        assert isinstance(transaction, Transaction)
        values = []
        result = transaction.run("UNWIND [1,2,3,4] AS x RETURN x")
        assert isinstance(result, Result)
        for record in result:
            values.append(record["x"])
        return values

    try:
        with GraphDatabase.driver(bolt_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = session.read_transaction(unwind)

        assert expected == [1, 2, 3, 4]
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_bolt_driver_multiple_results_case_1(bolt_uri, auth):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_bolt_driver_multiple_results_case_1
    try:
        with GraphDatabase.driver(bolt_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                transaction = session.begin_transaction(timeout=3, metadata={"foo": "bar"})
                result1 = transaction.run("UNWIND [1,2,3,4] AS x RETURN x")
                values1 = []
                for ix in result1:
                    values1.append(ix["x"])
                transaction.commit()
                assert values1 == [1, 2, 3, 4]

    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_bolt_driver_multiple_results_case_2(bolt_uri, auth):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_bolt_driver_multiple_results_case_2
    try:
        with GraphDatabase.driver(bolt_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                transaction = session.begin_transaction(timeout=3, metadata={"foo": "bar"})
                result1 = transaction.run("UNWIND [1,2,3,4] AS x RETURN x")
                result2 = transaction.run("UNWIND [5,6,7,8] AS x RETURN x")
                values1 = []
                values2 = []
                for ix in result2:
                    values2.append(ix["x"])
                for ix in result1:
                    values1.append(ix["x"])
                transaction.commit()
                assert values2 == [5, 6, 7, 8]
                assert values1 == [1, 2, 3, 4]

    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_bolt_driver_multiple_results_case_3(bolt_uri, auth):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_bolt_driver_multiple_results_case_3
    try:
        with GraphDatabase.driver(bolt_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                transaction = session.begin_transaction(timeout=3, metadata={"foo": "bar"})
                values1 = []
                values2 = []
                result1 = transaction.run("UNWIND [1,2,3,4] AS x RETURN x")
                result1_iter = iter(result1)
                values1.append(next(result1_iter)["x"])
                result2 = transaction.run("UNWIND [5,6,7,8] AS x RETURN x")
                for ix in result2:
                    values2.append(ix["x"])
                transaction.commit()  # Should discard the rest of records in result1 and result2 -> then set mode discard.
                assert values2 == [5, 6, 7, 8]
                assert result2._closed is True
                assert values1 == [1, ]

                try:
                    values1.append(next(result1_iter)["x"])
                except StopIteration as e:
                    # Bolt 4.0
                    assert values1 == [1, ]
                    assert result1._closed is True
                else:
                    # Bolt 3 only have PULL ALL and no qid so it will behave like autocommit r1=session.run rs2=session.run
                    values1.append(next(result1_iter)["x"])
                    assert values1 == [1, 2, 3]
                    assert result1._closed is False

        assert result1._closed is True

    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_bolt_driver_case_pull_no_records(driver):
    # python -m pytest tests/integration/test_bolt_driver.py -s -v -k test_bolt_driver_case_pull_no_records
    try:
        with driver.session(default_access_mode=WRITE_ACCESS) as session:
            with session.begin_transaction() as tx:
                result = tx.run("CREATE (a:Thing {uuid:$uuid})", uuid=123)
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


@fixture
def server_info(driver):
    """ Simple fixture to provide quick and easy access to a
    :class:`.ServerInfo` object.
    """
    with driver.session() as session:
        summary = session.run("RETURN 1").consume()
        yield summary.server


def test_server_address(server_info):
    assert server_info.address == ("127.0.0.1", 7687)


def test_server_protocol_version(server_info):
    assert server_info.protocol_version >= (3, 0)


def test_server_agent(server_info):
    assert server_info.agent.startswith("Neo4j/")


def test_server_connection_id(server_info):
    cid = server_info.connection_id
    assert cid.startswith("bolt-") and cid[5:].isdigit()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from neo4j.api import (
    kerberos_auth,
    basic_auth,
    custom_auth,
)

# python -m pytest -s -v tests/unit/test_security.py


def test_should_generate_kerberos_auth_token_correctly():
    auth = kerberos_auth("I am a base64 service ticket")
    assert auth.scheme == "kerberos"
    assert auth.principal == ""
    assert auth.credentials == "I am a base64 service ticket"
    assert not auth.realm
    assert not hasattr(auth, "parameters")


def test_should_generate_basic_auth_without_realm_correctly():
    auth = basic_auth("molly", "meoooow")
    assert auth.scheme == "basic"
    assert auth.principal == "molly"
    assert auth.credentials == "meoooow"
    assert not auth.realm
    assert not hasattr(auth, "parameters")


def test_should_generate_base_auth_with_realm_correctly():
    auth = basic_auth("molly", "meoooow", "cat_cafe")
    assert auth.scheme == "basic"
    assert auth.principal == "molly"
    assert auth.credentials == "meoooow"
    assert auth.realm == "cat_cafe"
    assert not hasattr(auth, "parameters")


def test_should_generate_custom_auth_correctly():
    auth = custom_auth("molly", "meoooow", "cat_cafe", "cat", age="1", color="white")
    assert auth.scheme == "cat"
    assert auth.principal == "molly"
    assert auth.credentials == "meoooow"
    assert auth.realm == "cat_cafe"
    assert auth.parameters == {"age": "1", "color": "white"}



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import pytest
from unittest.mock import NonCallableMagicMock

from neo4j import (
    ServerInfo,
    Session,
    SessionConfig,
)


class FakeConnection(NonCallableMagicMock):
    callbacks = []
    server_info = ServerInfo("127.0.0.1", (4, 3))

    def fetch_message(self, *args, **kwargs):
        if self.callbacks:
            cb = self.callbacks.pop(0)
            cb()
        return super().__getattr__("fetch_message")(*args, **kwargs)

    def fetch_all(self, *args, **kwargs):
        while self.callbacks:
            cb = self.callbacks.pop(0)
            cb()
        return super().__getattr__("fetch_all")(*args, **kwargs)

    def __getattr__(self, name):
        parent = super()

        def build_message_handler(name):
            def func(*args, **kwargs):
                def callback():
                    for cb_name, param_count in (
                        ("on_success", 1),
                        ("on_summary", 0)
                    ):
                        cb = kwargs.get(cb_name, None)
                        if callable(cb):
                            try:
                                param_count = \
                                    len(inspect.signature(cb).parameters)
                            except ValueError:
                                # e.g. built-in method as cb
                                pass
                            if param_count == 1:
                                cb({})
                            else:
                                cb()
                self.callbacks.append(callback)
                return parent.__getattr__(name)(*args, **kwargs)

            return func

        if name in ("run", "commit", "pull", "rollback", "discard"):
            return build_message_handler(name)
        return parent.__getattr__(name)

    def defunct(self):
        return False


@pytest.fixture()
def pool(mocker):
    pool = mocker.MagicMock()
    pool.acquire = mocker.MagicMock(side_effect=iter(FakeConnection, 0))
    return pool


def test_session_context_calls_close(mocker):
    s = Session(None, SessionConfig())
    mock_close = mocker.patch.object(s, 'close', autospec=True)
    with s:
        pass
    mock_close.assert_called_once_with()


@pytest.mark.parametrize("test_run_args", (
    ("RETURN $x", {"x": 1}), ("RETURN 1",)
))
@pytest.mark.parametrize(("repetitions", "consume"), (
    (1, False), (2, False), (2, True)
))
def test_opens_connection_on_run(pool, test_run_args, repetitions, consume):
    with Session(pool, SessionConfig()) as session:
        assert session._connection is None
        result = session.run(*test_run_args)
        assert session._connection is not None
        if consume:
            result.consume()


@pytest.mark.parametrize("test_run_args", (
    ("RETURN $x", {"x": 1}), ("RETURN 1",)
))
@pytest.mark.parametrize("repetitions", range(1, 3))
def test_closes_connection_after_consume(pool, test_run_args, repetitions):
    with Session(pool, SessionConfig()) as session:
        result = session.run(*test_run_args)
        result.consume()
        assert session._connection is None
    assert session._connection is None


@pytest.mark.parametrize("test_run_args", (
    ("RETURN $x", {"x": 1}), ("RETURN 1",)
))
def test_keeps_connection_until_last_result_consumed(pool, test_run_args):
    with Session(pool, SessionConfig()) as session:
        result1 = session.run(*test_run_args)
        result2 = session.run(*test_run_args)
        assert session._connection is not None
        result1.consume()
        assert session._connection is not None
        result2.consume()
        assert session._connection is None


def test_opens_connection_on_tx_begin(pool):
    with Session(pool, SessionConfig()) as session:
        assert session._connection is None
        with session.begin_transaction() as _:
            assert session._connection is not None


@pytest.mark.parametrize("test_run_args", (
    ("RETURN $x", {"x": 1}), ("RETURN 1",)
))
@pytest.mark.parametrize("repetitions", range(1, 3))
def test_keeps_connection_on_tx_run(pool, test_run_args, repetitions):
    with Session(pool, SessionConfig()) as session:
        with session.begin_transaction() as tx:
            for _ in range(repetitions):
                tx.run(*test_run_args)
                assert session._connection is not None


@pytest.mark.parametrize("test_run_args", (
        ("RETURN $x", {"x": 1}), ("RETURN 1",)
))
@pytest.mark.parametrize("repetitions", range(1, 3))
def test_keeps_connection_on_tx_consume(pool, test_run_args, repetitions):
    with Session(pool, SessionConfig()) as session:
        with session.begin_transaction() as tx:
            for _ in range(repetitions):
                result = tx.run(*test_run_args)
                result.consume()
                assert session._connection is not None


@pytest.mark.parametrize("test_run_args", (
        ("RETURN $x", {"x": 1}), ("RETURN 1",)
))
def test_closes_connection_after_tx_close(pool, test_run_args):
    with Session(pool, SessionConfig()) as session:
        with session.begin_transaction() as tx:
            for _ in range(2):
                result = tx.run(*test_run_args)
                result.consume()
            tx.close()
            assert session._connection is None
        assert session._connection is None


@pytest.mark.parametrize("test_run_args", (
        ("RETURN $x", {"x": 1}), ("RETURN 1",)
))
def test_closes_connection_after_tx_commit(pool, test_run_args):
    with Session(pool, SessionConfig()) as session:
        with session.begin_transaction() as tx:
            for _ in range(2):
                result = tx.run(*test_run_args)
                result.consume()
            tx.commit()
            assert session._connection is None
        assert session._connection is None


@pytest.mark.parametrize("bookmarks", (None, [], ["abc"], ["foo", "bar"]))
def test_session_returns_bookmark_directly(pool, bookmarks):
    with Session(pool, SessionConfig(bookmarks=bookmarks)) as session:
        if bookmarks:
            assert session.last_bookmark() == bookmarks[-1]
        else:
            assert session.last_bookmark() is None


@pytest.mark.parametrize(("query", "error_type"), (
    (None, ValueError),
    (1234, TypeError),
    ({"how about": "no?"}, TypeError),
    (["I don't", "think so"], TypeError),
))
def test_session_run_wrong_types(pool, query, error_type):
    with Session(pool, SessionConfig()) as session:
        with pytest.raises(error_type):
            session.run(query)
